@@ -4,12 +4,13 @@
 #include <rpm/rpmtag.h>
 #include <rpm/rpmstring.h>
 #include <rpm/rpmts.h>	/* XXX rpmtsCreate/rpmtsFree */
+#include <rpm/rpmver.h>
 
 #include "header-py.h"
 #include "rpmds-py.h"
 #include "rpmfd-py.h"
-#include "rpmfi-py.h"
 #include "rpmtd-py.h"
+#include "rpmver-py.h"
 
 /** \ingroup python
  * \class Rpm
@@ -143,7 +144,7 @@ static PyObject * hdrKeyList(hdrObject * s)
 
     hi = headerInitIterator(s->h);
     while ((tag = headerNextTag(hi)) != RPMTAG_NOT_FOUND) {
-	PyObject *to = PyInt_FromLong(tag);
+	PyObject *to = PyLong_FromLong(tag);
         if (!to) {
             headerFreeIterator(hi);
             Py_DECREF(keys);
@@ -161,10 +162,14 @@ static PyObject * hdrAsBytes(hdrObject * s)
 {
     PyObject *res = NULL;
     unsigned int len = 0;
+
     char *buf = headerExport(s->h, &len);
 
     if (buf == NULL || len == 0) {
-	PyErr_SetString(pyrpmError, "can't unload bad header\n");
+	rpmmodule_state_t *modstate = rpmModState_FromObject((PyObject*)s);
+	if (modstate) {
+	    PyErr_SetString(modstate->pyrpmError, "can't unload bad header\n");
+	}
     } else {
 	res = PyBytes_FromStringAndSize(buf, len);
     }
@@ -177,43 +182,6 @@ static PyObject * hdrUnload(hdrObject * s)
     return hdrAsBytes(s);
 }
 
-static PyObject * hdrExpandFilelist(hdrObject * s)
-{
-    DEPRECATED_METHOD("use hdr.convert() instead");
-    headerConvert(s->h, HEADERCONV_EXPANDFILELIST);
-
-    Py_RETURN_NONE;
-}
-
-static PyObject * hdrCompressFilelist(hdrObject * s)
-{
-    DEPRECATED_METHOD("use hdr.convert() instead");
-    headerConvert(s->h, HEADERCONV_COMPRESSFILELIST);
-
-    Py_RETURN_NONE;
-}
-
-/* make a header with _all_ the tags we need */
-static PyObject * hdrFullFilelist(hdrObject * s)
-{
-    rpmtd fileNames = rpmtdNew();
-    Header h = s->h;
-
-    DEPRECATED_METHOD("obsolete method");
-    if (!headerIsEntry (h, RPMTAG_BASENAMES)
-	|| !headerIsEntry (h, RPMTAG_DIRNAMES)
-	|| !headerIsEntry (h, RPMTAG_DIRINDEXES))
-	headerConvert(h, HEADERCONV_COMPRESSFILELIST);
-
-    if (headerGet(h, RPMTAG_FILENAMES, fileNames, HEADERGET_EXT)) {
-	rpmtdSetTag(fileNames, RPMTAG_OLDFILENAMES);
-	headerPut(h, fileNames, HEADERPUT_DEFAULT);
-    }
-    rpmtdFree(fileNames);
-
-    Py_RETURN_NONE;
-}
-
 static PyObject * hdrFormat(hdrObject * s, PyObject * args, PyObject * kwds)
 {
     const char * fmt;
@@ -227,7 +195,10 @@ static PyObject * hdrFormat(hdrObject * s, PyObject * args, PyObject * kwds)
 
     r = headerFormat(s->h, fmt, &err);
     if (!r) {
-	PyErr_SetString(pyrpmError, err);
+	rpmmodule_state_t *modstate = rpmModState_FromObject((PyObject*)s);
+	if (modstate) {
+	    PyErr_SetString(modstate->pyrpmError, err);
+	}
 	return NULL;
     }
 
@@ -265,12 +236,21 @@ static PyObject * hdrWrite(hdrObject *s, PyObject *args, PyObject *kwds)
 {
     char *kwlist[] = { "file", "magic", NULL };
     int magic = HEADER_MAGIC_YES;
+    PyObject *fdo_source;
     rpmfdObject *fdo = NULL;
     int rc;
-
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O&|i", kwlist,
-				     rpmfdFromPyObject, &fdo, &magic))
+    rpmmodule_state_t *modstate = rpmModState_FromObject((PyObject*)s);
+    if (!modstate) {
 	return NULL;
+    }
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|i", kwlist,
+				     &fdo_source, &magic))
+	return NULL;
+
+    if(!rpmfdFromPyObject(modstate, fdo_source, &fdo)) {
+	return NULL;
+    }
 
     Py_BEGIN_ALLOW_THREADS;
     rc = headerWrite(rpmfdGetFd(fdo), s->h,
@@ -284,38 +264,9 @@ static PyObject * hdrWrite(hdrObject *s, PyObject *args, PyObject *kwds)
     Py_RETURN_NONE;
 }
 
-/*
- * Just a backwards-compatibility dummy, the arguments are not looked at
- * or used. TODO: push this over to python side...
- */
-static PyObject * hdr_fiFromHeader(PyObject * s, PyObject * args, PyObject * kwds)
+static Py_hash_t hdr_hash(PyObject * h)
 {
-    return PyObject_CallFunctionObjArgs((PyObject *) &rpmfi_Type, s, NULL);
-}
-
-/* Backwards compatibility. Flags argument is just a dummy and discarded. */
-static PyObject * hdr_dsFromHeader(PyObject * s, PyObject * args, PyObject * kwds)
-{
-    rpmTagVal tag = RPMTAG_REQUIRENAME;
-    rpmsenseFlags flags = 0;
-    char * kwlist[] = {"to", "flags", NULL};
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|O&i:dsFromHeader", kwlist,
-            tagNumFromPyObject, &tag, &flags))
-        return NULL;
-
-    return PyObject_CallFunction((PyObject *) &rpmds_Type,
-                                 "(Oi)", s, tag);
-}
-
-static PyObject * hdr_dsOfHeader(PyObject * s)
-{
-    return PyObject_CallFunction((PyObject *) &rpmds_Type,
-                                 "(Oi)", s, RPMTAG_NEVR);
-}
-
-static long hdr_hash(PyObject * h)
-{
-    return (long) h;
+    return (Py_hash_t) h;
 }
 
 static PyObject * hdr_reduce(hdrObject *s)
@@ -334,28 +285,16 @@ static struct PyMethodDef hdr_methods[] = {
      "hdr.keys() -- Return a list of the header's rpm tags (int RPMTAG_*)." },
     {"unload",		(PyCFunction) hdrUnload,	METH_NOARGS,
      "hdr.unload() -- Return binary representation\nof the header." },
-    {"expandFilelist",	(PyCFunction) hdrExpandFilelist,METH_NOARGS,
-     "DEPRECATED -- Use hdr.convert() instead." },
-    {"compressFilelist",(PyCFunction) hdrCompressFilelist,METH_NOARGS,
-     "DEPRECATED -- Use hdr.convert() instead." },
-    {"fullFilelist",	(PyCFunction) hdrFullFilelist,	METH_NOARGS,
-     "DEPRECATED -- Obsolete method."},
     {"convert",		(PyCFunction) hdrConvert,	METH_VARARGS|METH_KEYWORDS,
      "hdr.convert(op=-1) -- Convert header - See HEADERCONV_*\nfor possible values of op."},
     {"format",		(PyCFunction) hdrFormat,	METH_VARARGS|METH_KEYWORDS,
-     "hdr.format(format) -- Expand a query string with the header data.\n\nSee rpm -q for syntax." },
+     "hdr.format(format) -- Expand an rpm-queryformat(7) string with the header data." },
     {"sprintf",		(PyCFunction) hdrFormat,	METH_VARARGS|METH_KEYWORDS,
      "Alias for .format()." },
     {"isSource",	(PyCFunction)hdrIsSource,	METH_NOARGS, 
      "hdr.isSource() -- Return if header describes a source package." },
     {"write",		(PyCFunction)hdrWrite,		METH_VARARGS|METH_KEYWORDS,
      "hdr.write(file, magic=True) -- Write header to file." },
-    {"dsOfHeader",	(PyCFunction)hdr_dsOfHeader,	METH_NOARGS,
-     "hdr.dsOfHeader() -- Return dependency set with the header's NEVR."},
-    {"dsFromHeader",	(PyCFunction)hdr_dsFromHeader,	METH_VARARGS|METH_KEYWORDS,
-     "hdr.dsFromHeader(to=RPMTAG_REQUIRENAME, flags=None)\nGet dependency set from header. to must be one of the NAME tags\nbelonging to a dependency:\n'Providename', 'Requirename', 'Obsoletename', 'Conflictname',\n'Triggername', 'Recommendname', 'Suggestname', 'Supplementname',\n'Enhancename' or one of the corresponding RPMTAG_*NAME constants." },
-    {"fiFromHeader",	(PyCFunction)hdr_fiFromHeader,	METH_VARARGS|METH_KEYWORDS,
-     "hdr.fiFromHeader() -- Return rpm.fi object containing the file\nmeta data from the header.\n\nDEPRECATED - Use rpm.files(hdr) instead."},
     {"__reduce__",	(PyCFunction)hdr_reduce,	METH_NOARGS,
 	NULL},
 
@@ -370,6 +309,11 @@ static PyObject *hdr_new(PyTypeObject *subtype, PyObject *args, PyObject *kwds)
     Header h = NULL;
     char *kwlist[] = { "obj", NULL };
 
+    rpmmodule_state_t *modstate = rpmModState_FromType(subtype);
+    if (!modstate) {
+	return NULL;
+    }
+
     if (!PyArg_ParseTupleAndKeywords(args, kwds, "|O", kwlist, &obj)) {
 	return NULL;
     }
@@ -379,8 +323,11 @@ static PyObject *hdr_new(PyTypeObject *subtype, PyObject *args, PyObject *kwds)
     } else if (hdrObject_Check(obj)) {
 	h = headerCopy(((hdrObject*) obj)->h);
     } else if (PyBytes_Check(obj)) {
-	h = headerCopyLoad(PyBytes_AsString(obj));
-    } else if (rpmfdFromPyObject(obj, &fdo)) {
+	Py_ssize_t len = 0;
+	char *blob = NULL;
+	if (PyBytes_AsStringAndSize(obj, &blob, &len) == 0)
+	    h = headerImport(blob, len, HEADERIMPORT_COPY);
+    } else if (rpmfdFromPyObject(modstate, obj, &fdo)) {
 	Py_BEGIN_ALLOW_THREADS;
 	h = headerRead(rpmfdGetFd(fdo), HEADER_MAGIC_YES);
 	Py_END_ALLOW_THREADS;
@@ -391,7 +338,7 @@ static PyObject *hdr_new(PyTypeObject *subtype, PyObject *args, PyObject *kwds)
     }
 
     if (h == NULL) {
-	PyErr_SetString(pyrpmError, "bad header");
+	PyErr_SetString(modstate->pyrpmError, "bad header");
 	return NULL;
     }
     
@@ -400,8 +347,20 @@ static PyObject *hdr_new(PyTypeObject *subtype, PyObject *args, PyObject *kwds)
 
 static void hdr_dealloc(hdrObject * s)
 {
+    PyObject_GC_UnTrack(s);
     if (s->h) headerFree(s->h);
-    Py_TYPE(s)->tp_free((PyObject *)s);
+    PyTypeObject *type = Py_TYPE(s);
+    freefunc free = PyType_GetSlot(type, Py_tp_free);
+    free(s);
+    Py_DECREF(type);
+}
+
+static int hdr_traverse(hdrObject * s, visitproc visit, void *arg)
+{
+    if (python_version >= 0x03090000) {
+        Py_VISIT(Py_TYPE(s));
+    }
+    return 0;
 }
 
 static PyObject * hdr_iternext(hdrObject *s)
@@ -412,7 +371,7 @@ static PyObject * hdr_iternext(hdrObject *s)
     if (s->hi == NULL) s->hi = headerInitIterator(s->h);
 
     if ((tag = headerNextTag(s->hi)) != RPMTAG_NOT_FOUND) {
-	res = PyInt_FromLong(tag);
+	res = PyLong_FromLong(tag);
     } else {
 	s->hi = headerFreeIterator(s->hi);
     }
@@ -421,14 +380,9 @@ static PyObject * hdr_iternext(hdrObject *s)
 
 PyObject * utf8FromString(const char *s)
 {
-/* In Python 3, we return all strings as surrogate-escaped utf-8 */
-#if PY_MAJOR_VERSION >= 3
+    /* We return all strings as surrogate-escaped utf-8 */
     if (s != NULL)
 	return PyUnicode_DecodeUTF8(s, strlen(s), "surrogateescape");
-#else
-    if (s != NULL)
-	return PyBytes_FromString(s);
-#endif
     Py_RETURN_NONE;
 }
 
@@ -452,9 +406,9 @@ int tagNumFromPyObject (PyObject *item, rpmTagVal *tagp)
     rpmTagVal tag = RPMTAG_NOT_FOUND;
     PyObject *str = NULL;
 
-    if (PyInt_Check(item)) {
+    if (PyLong_Check(item)) {
 	/* XXX we should probably validate tag numbers too */
-	tag = PyInt_AsLong(item);
+	tag = PyLong_AsLong(item);
     } else if (utf8FromPyObject(item, &str)) {
 	tag = rpmTagGetValue(PyBytes_AsString(str));
 	Py_DECREF(str);
@@ -471,14 +425,14 @@ int tagNumFromPyObject (PyObject *item, rpmTagVal *tagp)
     return 1;
 }
 
-static PyObject * hdrGetTag(Header h, rpmTagVal tag)
+static PyObject * hdrGetTag(rpmmodule_state_t *modstate, Header h, rpmTagVal tag)
 {
     PyObject *res = NULL;
     struct rpmtd_s td;
 
     (void) headerGet(h, tag, &td, HEADERGET_EXT);
     if (rpmtdGetFlags(&td) & RPMTD_INVALID) {
-	PyErr_SetString(pyrpmError, "invalid header data");
+	PyErr_SetString(modstate->pyrpmError, "invalid header data");
     } else {
 	/* rpmtd_AsPyobj() knows how to handle empty containers and all */
 	res = rpmtd_AsPyobj(&td);
@@ -493,7 +447,7 @@ static int validItem(rpmTagClass tclass, PyObject *item)
 
     switch (tclass) {
     case RPM_NUMERIC_CLASS:
-	rc = (PyLong_Check(item) || PyInt_Check(item));
+	rc = PyLong_Check(item);
 	break;
     case RPM_STRING_CLASS:
 	rc = (PyBytes_Check(item) || PyUnicode_Check(item));
@@ -552,20 +506,20 @@ static int hdrAppendItem(Header h, rpmTagVal tag, rpmTagType type, PyObject *ite
 	rc = headerPutBin(h, tag, val, len);
 	} break;
     case RPM_INT64_TYPE: {
-	uint64_t val = PyInt_AsUnsignedLongLongMask(item);
+	uint64_t val = PyLong_AsUnsignedLongLongMask(item);
 	rc = headerPutUint64(h, tag, &val, 1);
 	} break;
     case RPM_INT32_TYPE: {
-	uint32_t val = PyInt_AsUnsignedLongMask(item);
+	uint32_t val = PyLong_AsUnsignedLongMask(item);
 	rc = headerPutUint32(h, tag, &val, 1);
 	} break;
     case RPM_INT16_TYPE: {
-	uint16_t val = PyInt_AsUnsignedLongMask(item);
+	uint16_t val = PyLong_AsUnsignedLongMask(item);
 	rc = headerPutUint16(h, tag, &val, 1);
 	} break;
     case RPM_INT8_TYPE:
     case RPM_CHAR_TYPE: {
-	uint8_t val = PyInt_AsUnsignedLongMask(item);
+	uint8_t val = PyLong_AsUnsignedLongMask(item);
 	rc = headerPutUint8(h, tag, &val, 1);
 	} break;
     default:
@@ -611,8 +565,13 @@ static PyObject * hdr_subscript(hdrObject * s, PyObject * item)
 {
     rpmTagVal tag;
 
+    rpmmodule_state_t *modstate = rpmModState_FromObject((PyObject*)s);
+    if (!modstate) {
+	return NULL;
+    }
+
     if (!tagNumFromPyObject(item, &tag)) return NULL;
-    return hdrGetTag(s->h, tag);
+    return hdrGetTag(modstate, s->h, tag);
 }
 
 static int hdr_ass_subscript(hdrObject *s, PyObject *key, PyObject *value)
@@ -620,6 +579,11 @@ static int hdr_ass_subscript(hdrObject *s, PyObject *key, PyObject *value)
     rpmTagVal tag;
     if (!tagNumFromPyObject(key, &tag)) return -1;
 
+    rpmmodule_state_t *modstate = rpmModState_FromObject((PyObject*)s);
+    if (!modstate) {
+	return -1;
+    }
+
     if (value == NULL) {
 	/* XXX should failure raise key error? */
 	headerDel(s->h, tag);
@@ -631,6 +595,11 @@ static int hdr_ass_subscript(hdrObject *s, PyObject *key, PyObject *value)
 
 static PyObject * hdr_getattro(hdrObject * s, PyObject * n)
 {
+    rpmmodule_state_t *modstate = rpmModState_FromObject((PyObject*)s);
+    if (!modstate) {
+	return NULL;
+    }
+
     PyObject *res = PyObject_GenericGetAttr((PyObject *) s, n);
     if (res == NULL) {
 	PyObject *type, *value, *traceback;
@@ -642,7 +611,7 @@ static PyObject * hdr_getattro(hdrObject * s, PyObject * n)
 	    Py_XDECREF(type);
 	    Py_XDECREF(value);
 	    Py_XDECREF(traceback);
-	    res = hdrGetTag(s->h, tag);
+	    res = hdrGetTag(modstate, s->h, tag);
 	} else {
 	    PyErr_Restore(type, value, traceback);
 	}
@@ -655,24 +624,7 @@ static int hdr_setattro(PyObject * o, PyObject * n, PyObject * v)
     return PyObject_GenericSetAttr(o, n, v);
 }
 
-static PyMappingMethods hdr_as_mapping = {
-	(lenfunc) 0,			/* mp_length */
-	(binaryfunc) hdr_subscript,	/* mp_subscript */
-	(objobjargproc)hdr_ass_subscript,/* mp_ass_subscript */
-};
 
-static PySequenceMethods hdr_as_sequence = {
-    0,				/* sq_length */
-    0,				/* sq_concat */
-    0,				/* sq_repeat */
-    0,				/* sq_item */
-    0,				/* sq_slice */
-    0,				/* sq_ass_item */
-    0,				/* sq_ass_slice */
-    (objobjproc)hdrContains,	/* sq_contains */
-    0,				/* sq_inplace_concat */
-    0,				/* sq_inplace_repeat */
-};
 
 static char hdr_doc[] =
   "A header object represents an RPM package header.\n"
@@ -726,52 +678,33 @@ static char hdr_doc[] =
   "the strings in header lookups don't get translated, or the lookups\n"
   "will fail.\n";
 
-PyTypeObject hdr_Type = {
-	PyVarObject_HEAD_INIT(&PyType_Type, 0)
-	"rpm.hdr",			/* tp_name */
-	sizeof(hdrObject),		/* tp_size */
-	0,				/* tp_itemsize */
-	(destructor) hdr_dealloc, 	/* tp_dealloc */
-	0,				/* tp_print */
-	(getattrfunc) 0, 		/* tp_getattr */
-	0,				/* tp_setattr */
-	0,				/* tp_compare */
-	0,				/* tp_repr */
-	0,				/* tp_as_number */
-	&hdr_as_sequence,		/* tp_as_sequence */
-	&hdr_as_mapping,		/* tp_as_mapping */
-	hdr_hash,			/* tp_hash */
-	0,				/* tp_call */
-	0,				/* tp_str */
-	(getattrofunc) hdr_getattro,	/* tp_getattro */
-	(setattrofunc) hdr_setattro,	/* tp_setattro */
-	0,				/* tp_as_buffer */
-	Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,	/* tp_flags */
-	hdr_doc,			/* tp_doc */
-	0,				/* tp_traverse */
-	0,				/* tp_clear */
-	0,				/* tp_richcompare */
-	0,				/* tp_weaklistoffset */
-	PyObject_SelfIter,		/* tp_iter */
-	(iternextfunc)hdr_iternext,	/* tp_iternext */
-	hdr_methods,			/* tp_methods */
-	0,				/* tp_members */
-	0,				/* tp_getset */
-	0,				/* tp_base */
-	0,				/* tp_dict */
-	0,				/* tp_descr_get */
-	0,				/* tp_descr_set */
-	0,				/* tp_dictoffset */
-	0,				/* tp_init */
-	0,				/* tp_alloc */
-	hdr_new,			/* tp_new */
-	0,				/* tp_free */
-	0,				/* tp_is_gc */
+static PyType_Slot hdr_Type_Slots[] = {
+    {Py_tp_dealloc, hdr_dealloc},
+    {Py_tp_traverse, hdr_traverse},
+    {Py_sq_contains, hdrContains},
+    {Py_mp_subscript, hdr_subscript},
+    {Py_mp_ass_subscript, hdr_ass_subscript},
+    {Py_tp_hash, hdr_hash},
+    {Py_tp_getattro, hdr_getattro},
+    {Py_tp_setattro, hdr_setattro},
+    {Py_tp_doc, hdr_doc},
+    {Py_tp_iter, PyObject_SelfIter},
+    {Py_tp_iternext, hdr_iternext},
+    {Py_tp_methods, hdr_methods},
+    {Py_tp_new, hdr_new},
+    {0, NULL},
+};
+PyType_Spec hdr_Type_Spec = {
+    .name = "rpm.hdr",
+    .basicsize = sizeof(hdrObject),
+    .flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_IMMUTABLETYPE,
+    .slots = hdr_Type_Slots,
 };
 
 PyObject * hdr_Wrap(PyTypeObject *subtype, Header h)
 {
-    hdrObject * hdr = (hdrObject *)subtype->tp_alloc(subtype, 0);
+    allocfunc subtype_alloc = (allocfunc)PyType_GetSlot(subtype, Py_tp_alloc);
+    hdrObject *hdr = (hdrObject *)subtype_alloc(subtype, 0);
     if (hdr == NULL) return NULL;
     hdr->h = h;
     return (PyObject *) hdr;
@@ -788,141 +721,38 @@ int hdrFromPyObject(PyObject *item, Header *hptr)
     }
 }
 
-/**
- * This assumes the order of list matches the order of the new headers, and
- * throws an exception if that isn't true.
- */
-static int rpmMergeHeaders(PyObject * list, FD_t fd, int matchTag)
-{
-    Header h;
-    HeaderIterator hi;
-    rpmTagVal newMatch, oldMatch;
-    hdrObject * hdr;
-    rpm_count_t count = 0;
-    int rc = 1; /* assume failure */
-    rpmtd td = rpmtdNew();
-
-    Py_BEGIN_ALLOW_THREADS
-    h = headerRead(fd, HEADER_MAGIC_YES);
-    Py_END_ALLOW_THREADS
-
-    while (h) {
-	if (!headerGet(h, matchTag, td, HEADERGET_MINMEM)) {
-	    PyErr_SetString(pyrpmError, "match tag missing in new header");
-	    goto exit;
-	}
-	newMatch = rpmtdTag(td);
-	rpmtdFreeData(td);
-
-	hdr = (hdrObject *) PyList_GetItem(list, count++);
-	if (!hdr) goto exit;
-
-	if (!headerGet(hdr->h, matchTag, td, HEADERGET_MINMEM)) {
-	    PyErr_SetString(pyrpmError, "match tag missing in new header");
-	    goto exit;
-	}
-	oldMatch = rpmtdTag(td);
-	rpmtdFreeData(td);
-
-	if (newMatch != oldMatch) {
-	    PyErr_SetString(pyrpmError, "match tag mismatch");
-	    goto exit;
-	}
-
-	for (hi = headerInitIterator(h); headerNext(hi, td); rpmtdFreeData(td))
-	{
-	    /* could be dupes */
-	    headerDel(hdr->h, rpmtdTag(td));
-	    headerPut(hdr->h, td, HEADERPUT_DEFAULT);
-	}
-
-	headerFreeIterator(hi);
-	h = headerFree(h);
-
-	Py_BEGIN_ALLOW_THREADS
-	h = headerRead(fd, HEADER_MAGIC_YES);
-	Py_END_ALLOW_THREADS
-    }
-    rc = 0;
-
-exit:
-    rpmtdFree(td);
-    return rc;
-}
-
-PyObject *
-rpmMergeHeadersFromFD(PyObject * self, PyObject * args, PyObject * kwds)
-{
-    FD_t fd;
-    int fileno;
-    PyObject * list;
-    int rc;
-    int matchTag;
-    char * kwlist[] = {"list", "fd", "matchTag", NULL};
-
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "Oii", kwlist, &list,
-	    &fileno, &matchTag))
-	return NULL;
-
-    if (!PyList_Check(list)) {
-	PyErr_SetString(PyExc_TypeError, "first parameter must be a list");
-	return NULL;
-    }
-
-    fd = fdDup(fileno);
-
-    rc = rpmMergeHeaders (list, fd, matchTag);
-    Fclose(fd);
-
-    if (rc) {
-	return NULL;
-    }
-
-    Py_RETURN_NONE;
-}
-
-PyObject * versionCompare (PyObject * self, PyObject * args, PyObject * kwds)
+PyObject * versionCompare (PyObject *mod, PyObject * args, PyObject * kwds)
 {
     hdrObject * h1, * h2;
     char * kwlist[] = {"version0", "version1", NULL};
 
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O!O!", kwlist, &hdr_Type,
-	    &h1, &hdr_Type, &h2))
+    rpmmodule_state_t *modstate = rpmModState_FromModule((PyObject*)mod);
+    if (!modstate) {
+	return NULL;
+    }
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O!O!", kwlist, modstate->hdr_Type,
+                                     &h1, modstate->hdr_Type, &h2))
 	return NULL;
 
     return Py_BuildValue("i", rpmVersionCompare(h1->h, h2->h));
 }
 
-static int compare_values(const char *str1, const char *str2)
+PyObject * labelCompare (PyObject *mod, PyObject * args)
 {
-    if (!str1 && !str2)
-	return 0;
-    else if (str1 && !str2)
-	return 1;
-    else if (!str1 && str2)
-	return -1;
-    return rpmvercmp(str1, str2);
-}
+    PyObject *rco = NULL;
+    rpmver rv1 = NULL;
+    rpmver rv2 = NULL;
 
-PyObject * labelCompare (PyObject * self, PyObject * args)
-{
-    const char *v1, *r1, *v2, *r2;
-    const char *e1, *e2;
-    int rc;
-
-    if (!PyArg_ParseTuple(args, "(zzz)(zzz)",
-			&e1, &v1, &r1, &e2, &v2, &r2))
+    if (!PyArg_ParseTuple(args, "O&O&",
+			    verFromPyObject, &rv1, verFromPyObject, &rv2)) {
 	return NULL;
-
-    if (e1 == NULL)	e1 = "0";
-    if (e2 == NULL)	e2 = "0";
-
-    rc = compare_values(e1, e2);
-    if (!rc) {
-	rc = compare_values(v1, v2);
-	if (!rc)
-	    rc = compare_values(r1, r2);
     }
-    return Py_BuildValue("i", rc);
+
+    rco = Py_BuildValue("i", rpmverCmp(rv1, rv2));
+
+    rpmverFree(rv1);
+    rpmverFree(rv2);
+
+    return rco;
 }
 
