@@ -12,7 +12,6 @@
 #include "rpmarchive-py.h"
 #include "rpmds-py.h"
 #include "rpmfd-py.h"
-#include "rpmfi-py.h"
 #include "rpmfiles-py.h"
 #include "rpmkeyring-py.h"
 #include "rpmmi-py.h"
@@ -23,13 +22,14 @@
 #include "rpmtd-py.h"
 #include "rpmte-py.h"
 #include "rpmts-py.h"
+#include "rpmver-py.h"
 #include "spec-py.h"
 
 /** \ingroup python
  * \name Module: rpm
  */
 
-PyObject * pyrpmError;
+unsigned long python_version = 0;
 
 static PyObject * archScore(PyObject * self, PyObject * arg)
 {
@@ -41,19 +41,6 @@ static PyObject * archScore(PyObject * self, PyObject * arg)
     return Py_BuildValue("i", rpmMachineScore(RPM_MACHTABLE_INSTARCH, arch));
 }
 
-static PyObject * signalCaught(PyObject *self, PyObject *o)
-{
-    int signo;
-    if (!PyArg_Parse(o, "i", &signo)) return NULL;
-
-    return PyBool_FromLong(rpmsqIsCaught(signo));
-}
-
-static PyObject * checkSignals(PyObject * self)
-{
-    return Py_BuildValue("i", rpmsqPoll());
-}
-
 static PyObject * blockSignals(PyObject * self, PyObject *arg)
 {
     int block;
@@ -99,15 +86,6 @@ setVerbosity (PyObject * self, PyObject * arg)
     Py_RETURN_NONE;
 }
 
-static PyObject *
-setEpochPromote (PyObject * self, PyObject * arg)
-{
-    if (!PyArg_Parse(arg, "i", &_rpmds_nopromote))
-	return NULL;
-
-    Py_RETURN_NONE;
-}
-
 static PyObject * setStats (PyObject * self, PyObject * arg)
 {
     if (!PyArg_Parse(arg, "i", &_rpmts_stats))
@@ -144,20 +122,6 @@ static PyObject * reloadConfig(PyObject * self, PyObject * args, PyObject *kwds)
     return PyBool_FromLong(rc == 0);
 }
 
-static PyObject * setInterruptSafety(PyObject * self, PyObject * args, PyObject *kwds)
-{
-    int on = 1;
-    PyObject * obj;
-    char * kwlist[] = { "on", NULL };
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|O", kwlist, &obj))
-	return NULL;
-    if (obj) {
-	on = PyObject_IsTrue(obj);
-    }
-    rpmsqSetInterruptSafety(on);
-    Py_RETURN_NONE;
-}
-
 static int parseSignArgs(PyObject * args, PyObject *kwds,
 			const char **path, struct rpmSignArgs *sargs)
 {
@@ -205,16 +169,9 @@ static PyMethodDef rpmModuleMethods[] = {
       "archscore(archname) -- How well does an architecture fit on this machine\n\n"
       "0 for non matching arch names\n1 for best arch\nhigher numbers for less fitting arches\n(e.g. 2 for \"i586\" on an i686 machine)" },
 
-    { "signalCaught", (PyCFunction) signalCaught, METH_O, 
-	"signalCaught(signo) -- Returns True if signal was caught." },
-    { "checkSignals", (PyCFunction) checkSignals, METH_NOARGS,
-      "checkSignals() -- Check for and exit on termination signals."},
     { "blockSignals", (PyCFunction) blockSignals, METH_O,
       "blocksignals(True/False) -- Block/unblock signals, refcounted."},
 
-    { "mergeHeaderListFromFD", (PyCFunction) rpmMergeHeadersFromFD, METH_VARARGS|METH_KEYWORDS,
-	NULL },
-
     { "log",		(PyCFunction) doLog, METH_VARARGS|METH_KEYWORDS,
 	"log(level, msg) -- Write msg to log if level is selected to be logged.\n\n"
     "level must be one of the RPMLOG_* constants."},
@@ -231,24 +188,11 @@ static PyMethodDef rpmModuleMethods[] = {
       "but arguments are tuples of of strings for (epoch, version, release)"},
     { "setVerbosity", (PyCFunction) setVerbosity, METH_O,
       "setVerbosity(level) -- Set log level. See RPMLOG_* constants." },
-    { "setEpochPromote", (PyCFunction) setEpochPromote, METH_O,
-	"setEpochPromote(bool) -- Set if no epoch shall be treated as epoch 0" },
     { "setStats", (PyCFunction) setStats, METH_O,
       "setStats(bool) -- Set if timing stats are printed after a transaction."},
     { "reloadConfig", (PyCFunction) reloadConfig, METH_VARARGS|METH_KEYWORDS,
       "reloadConfig(target=None) -- Reload config from files.\n\n"
       "Set all macros and settings accordingly."},
-
-    { "setInterruptSafety", (PyCFunction) setInterruptSafety,
-      METH_VARARGS|METH_KEYWORDS,
-      "setInterruptSafety(on=True) -- Set if various signals get intercepted.\n\n"
-      "By default, librpm will trap various unix signals (like SIGINT and\n"
-      "SIGTERM), in order to avoid process exit while locks are held or\n"
-      "a transaction is being performed.\n\n"
-      "If this is not the desired behaviour it's recommended to call this\n"
-      "once only at process startup because currently signal handlers will\n"
-      "not be retroactively applied if a database is open."
-    },
     { "addSign", (PyCFunction) addSign, METH_VARARGS|METH_KEYWORDS, NULL },
     { "delSign", (PyCFunction) delSign, METH_VARARGS|METH_KEYWORDS, NULL },
     { NULL }
@@ -272,7 +216,7 @@ static void addRpmTags(PyObject *module)
 	tagval = rpmTagGetValue(shortname);
 
 	PyModule_AddIntConstant(module, tagname, tagval);
-	pyval = PyInt_FromLong(tagval);
+	pyval = PyLong_FromLong(tagval);
 	pyname = utf8FromString(shortname);
 	PyDict_SetItem(dict, pyval, pyname);
 	Py_DECREF(pyval);
@@ -282,146 +226,221 @@ static void addRpmTags(PyObject *module)
     rpmtdFree(names);
 }
 
-/*
-  Do any common preliminary work before python 2 vs python 3 module creation:
-*/
-static int prepareInitModule(void)
-{
-    if (PyType_Ready(&hdr_Type) < 0) return 0;
-    if (PyType_Ready(&rpmarchive_Type) < 0) return 0;
-    if (PyType_Ready(&rpmds_Type) < 0) return 0;
-    if (PyType_Ready(&rpmfd_Type) < 0) return 0;
-    if (PyType_Ready(&rpmfi_Type) < 0) return 0;
-    if (PyType_Ready(&rpmfile_Type) < 0) return 0;
-    if (PyType_Ready(&rpmfiles_Type) < 0) return 0;
-    if (PyType_Ready(&rpmKeyring_Type) < 0) return 0;
-    if (PyType_Ready(&rpmmi_Type) < 0) return 0;
-    if (PyType_Ready(&rpmii_Type) < 0) return 0;
-    if (PyType_Ready(&rpmProblem_Type) < 0) return 0;
-    if (PyType_Ready(&rpmPubkey_Type) < 0) return 0;
-    if (PyType_Ready(&rpmstrPool_Type) < 0) return 0;
-    if (PyType_Ready(&rpmte_Type) < 0) return 0;
-    if (PyType_Ready(&rpmts_Type) < 0) return 0;
-    if (PyType_Ready(&spec_Type) < 0) return 0;
-    if (PyType_Ready(&specPkg_Type) < 0) return 0;
-
-    return 1;
-}
 static int initModule(PyObject *m);
 
-#if PY_MAJOR_VERSION >= 3
 static int rpmModuleTraverse(PyObject *m, visitproc visit, void *arg) {
-    Py_VISIT(pyrpmError);
+    rpmmodule_state_t *modstate = PyModule_GetState(m);
+    Py_VISIT(modstate->hdr_Type);
+    Py_VISIT(modstate->rpmarchive_Type);
+    Py_VISIT(modstate->rpmds_Type);
+    Py_VISIT(modstate->rpmfd_Type);
+    Py_VISIT(modstate->rpmfile_Type);
+    Py_VISIT(modstate->rpmfiles_Type);
+    Py_VISIT(modstate->rpmii_Type);
+    Py_VISIT(modstate->rpmKeyring_Type);
+    Py_VISIT(modstate->rpmPubkey_Type);
+    Py_VISIT(modstate->rpmmi_Type);
+    Py_VISIT(modstate->rpmProblem_Type);
+    Py_VISIT(modstate->rpmstrPool_Type);
+    Py_VISIT(modstate->rpmte_Type);
+    Py_VISIT(modstate->rpmts_Type);
+    Py_VISIT(modstate->rpmver_Type);
+    Py_VISIT(modstate->spec_Type);
+    Py_VISIT(modstate->specPkg_Type);
+    Py_VISIT(modstate->pyrpmError);
     return 0;
 }
 
 static int rpmModuleClear(PyObject *m) {
-    Py_CLEAR(pyrpmError);
+    rpmmodule_state_t *modstate = PyModule_GetState(m);
+    Py_CLEAR(modstate->hdr_Type);
+    Py_CLEAR(modstate->rpmarchive_Type);
+    Py_CLEAR(modstate->rpmds_Type);
+    Py_CLEAR(modstate->rpmfd_Type);
+    Py_CLEAR(modstate->rpmfile_Type);
+    Py_CLEAR(modstate->rpmfiles_Type);
+    Py_CLEAR(modstate->rpmii_Type);
+    Py_CLEAR(modstate->rpmKeyring_Type);
+    Py_CLEAR(modstate->rpmPubkey_Type);
+    Py_CLEAR(modstate->rpmmi_Type);
+    Py_CLEAR(modstate->rpmProblem_Type);
+    Py_CLEAR(modstate->rpmstrPool_Type);
+    Py_CLEAR(modstate->rpmte_Type);
+    Py_CLEAR(modstate->rpmts_Type);
+    Py_CLEAR(modstate->rpmver_Type);
+    Py_CLEAR(modstate->spec_Type);
+    Py_CLEAR(modstate->specPkg_Type);
+    Py_CLEAR(modstate->pyrpmError);
     return 0;
 }
 
-static struct PyModuleDef moduledef = {
-    PyModuleDef_HEAD_INIT,
-    "_rpm",            /* m_name */
-    rpm__doc__,        /* m_doc */
-    0,                 /* m_size */
-    rpmModuleMethods,
-    NULL,              /* m_reload */
-    rpmModuleTraverse,
-    rpmModuleClear,
-    NULL               /* m_free */
+static void rpmModuleFree(void *m) {
+    (void)rpmModuleClear(m);
+}
+
+static PyModuleDef_Slot module_slots[] = {
+    {Py_mod_exec, initModule},
+    {0}
 };
 
+static struct PyModuleDef moduledef = {
+    PyModuleDef_HEAD_INIT,
+    "_rpm",			/* m_name */
+    rpm__doc__,			/* m_doc */
+    sizeof(rpmmodule_state_t),	/* m_size */
+    rpmModuleMethods,
+    module_slots,      /* m_slots */
+    rpmModuleTraverse,
+    rpmModuleClear,
+    rpmModuleFree
+};
+
+/* The init function must be exported as it's called by Python, but it doesn't
+ * need a declaration in a header file.
+ * Provide a declaration to avoid -Wmissing-prototypes warnings.
+ */
 PyObject *
 PyInit__rpm(void);
 
 PyObject *
 PyInit__rpm(void)
 {
-    PyObject * m;
-    if (!prepareInitModule()) return NULL;
-    m = PyModule_Create(&moduledef);
-    initModule(m);
-    return m;
+    return PyModuleDef_Init(&moduledef);
 }
-#else
-void init_rpm(void);	/* XXX eliminate gcc warning */
-void init_rpm(void)
+
+/* Create a type object based on a Spec, and add it to the module. */
+static int initAndAddType(PyObject *m, PyTypeObject **type, PyType_Spec *spec,
+                          char *name)
 {
-    PyObject * m;
-
-    if (!prepareInitModule()) return;
-    m = Py_InitModule3("_rpm", rpmModuleMethods, rpm__doc__);
-    if (m == NULL)
-	return;
-    initModule(m);
+    if (!*type) {
+        *type = (PyTypeObject *)PyType_FromModuleAndSpec(m, spec, NULL);
+        if (!*type) return 0;
+    }
+    /* We intentionally leak a reference to `type` (only once per type per
+     * process).
+     */
+    Py_INCREF(*type);
+    /* Reference counting for PyModule_AddObject is tricky (see
+     * PyModule_AddObject docs). But let's do it right, as if we haven't just
+     * leaked.
+     * (Simpler API, `PyModule_AddObjectRef`, is only in Python 3.10+.)
+     */
+    if (PyModule_AddObject(m, name, (PyObject *) *type) < 0) {
+        Py_DECREF(*type);
+        return 0;
+    }
+    return 1;
 }
-#endif
 
-/* Shared python2/3 module initialization: */
+static unsigned long _get_python_version(void) {
+    PyObject *python_version_tuple = PySys_GetObject("version_info");
+    if (python_version_tuple == NULL) {
+        PyErr_SetString(PyExc_SystemError, "could not get Python version");
+	return 0;
+    }
+    unsigned int major, minor, micro, serial;
+    PyObject *releaselevel;
+    if (!PyArg_ParseTuple(
+	    python_version_tuple, "IIIOI",
+	    &major, &minor, &micro, &releaselevel, &serial)) {
+	return 0;
+    }
+    return (major << 24) | (minor << 16) | (micro << 8);
+}
+
+/* Module initialization: */
 static int initModule(PyObject *m)
 {
     PyObject * d;
 
     /* failure to initialize rpm (crypto and all) is rather fatal too... */
     if (rpmReadConfigFiles(NULL, NULL) == -1)
-	return 0;
+	return -1;
 
     d = PyModule_GetDict(m);
 
-    pyrpmError = PyErr_NewException("_rpm.error", NULL, NULL);
-    if (pyrpmError != NULL)
-	PyDict_SetItemString(d, "error", pyrpmError);
+    if (python_version == 0) {
+	python_version = _get_python_version();
+	if (python_version == 0) {
+	return -1;
+	}
+    }
 
-    Py_INCREF(&hdr_Type);
-    PyModule_AddObject(m, "hdr", (PyObject *) &hdr_Type);
+    rpmmodule_state_t *modstate = PyModule_GetState(m);
+    if (!modstate) {
+	return -1;
+    }
 
-    Py_INCREF(&rpmarchive_Type);
-    PyModule_AddObject(m, "archive", (PyObject *) &rpmarchive_Type);
+    modstate->pyrpmError = PyErr_NewException("_rpm.error", NULL, NULL);
+    if (modstate->pyrpmError != NULL)
+	PyDict_SetItemString(d, "error", modstate->pyrpmError);
 
-    Py_INCREF(&rpmds_Type);
-    PyModule_AddObject(m, "ds", (PyObject *) &rpmds_Type);
+    if (!initAndAddType(m, &modstate->hdr_Type, &hdr_Type_Spec, "hdr")) {
+	return -1;
+    }
 
-    Py_INCREF(&rpmfd_Type);
-    PyModule_AddObject(m, "fd", (PyObject *) &rpmfd_Type);
+    if (!initAndAddType(m, &modstate->rpmarchive_Type, &rpmarchive_Type_Spec, "archive")) {
+	return -1;
+    }
 
-    Py_INCREF(&rpmfi_Type);
-    PyModule_AddObject(m, "fi", (PyObject *) &rpmfi_Type);
+    if (!initAndAddType(m, &modstate->rpmds_Type, &rpmds_Type_Spec, "ds")) {
+	return -1;
+    }
 
-    Py_INCREF(&rpmfile_Type);
-    PyModule_AddObject(m, "file", (PyObject *) &rpmfile_Type);
+    if (!initAndAddType(m, &modstate->rpmfd_Type, &rpmfd_Type_Spec, "fd")) {
+	return -1;
+    }
 
-    Py_INCREF(&rpmfiles_Type);
-    PyModule_AddObject(m, "files", (PyObject *) &rpmfiles_Type);
+    if (!initAndAddType(m, &modstate->rpmfile_Type, &rpmfile_Type_Spec, "file")) {
+	return -1;
+    }
 
-    Py_INCREF(&rpmKeyring_Type);
-    PyModule_AddObject(m, "keyring", (PyObject *) &rpmKeyring_Type);
+    if (!initAndAddType(m, &modstate->rpmfiles_Type, &rpmfiles_Type_Spec, "files")) {
+	return -1;
+    }
 
-    Py_INCREF(&rpmmi_Type);
-    PyModule_AddObject(m, "mi", (PyObject *) &rpmmi_Type);
+    if (!initAndAddType(m, &modstate->rpmKeyring_Type, &rpmKeyring_Type_Spec, "keyring")) {
+	return -1;
+    }
 
-    Py_INCREF(&rpmii_Type);
-    PyModule_AddObject(m, "ii", (PyObject *) &rpmii_Type);
+    if (!initAndAddType(m, &modstate->rpmmi_Type, &rpmmi_Type_Spec, "mi")) {
+	return -1;
+    }
 
-    Py_INCREF(&rpmProblem_Type);
-    PyModule_AddObject(m, "prob", (PyObject *) &rpmProblem_Type);
+    if (!initAndAddType(m, &modstate->rpmii_Type, &rpmii_Type_Spec, "ii")) {
+	return -1;
+    }
 
-    Py_INCREF(&rpmPubkey_Type);
-    PyModule_AddObject(m, "pubkey", (PyObject *) &rpmPubkey_Type);
+    if (!initAndAddType(m, &modstate->rpmProblem_Type, &rpmProblem_Type_Spec, "prob")) {
+	return -1;
+    }
 
-    Py_INCREF(&rpmstrPool_Type);
-    PyModule_AddObject(m, "strpool", (PyObject *) &rpmstrPool_Type);
+    if (!initAndAddType(m, &modstate->rpmPubkey_Type, &rpmPubkey_Type_Spec, "pubkey")) {
+	return -1;
+    }
 
-    Py_INCREF(&rpmte_Type);
-    PyModule_AddObject(m, "te", (PyObject *) &rpmte_Type);
+    if (!initAndAddType(m, &modstate->rpmstrPool_Type, &rpmstrPool_Type_Spec, "strpool")) {
+	return -1;
+    }
 
-    Py_INCREF(&rpmts_Type);
-    PyModule_AddObject(m, "ts", (PyObject *) &rpmts_Type);
+    if (!initAndAddType(m, &modstate->rpmte_Type, &rpmte_Type_Spec, "te")) {
+	return -1;
+    }
 
-    Py_INCREF(&spec_Type);
-    PyModule_AddObject(m, "spec", (PyObject *) &spec_Type);
-    Py_INCREF(&specPkg_Type);
-    PyModule_AddObject(m, "specPkg", (PyObject *) &specPkg_Type);
+    if (!initAndAddType(m, &modstate->rpmts_Type, &rpmts_Type_Spec, "ts")) {
+	return -1;
+    }
+
+    if (!initAndAddType(m, &modstate->rpmver_Type, &rpmver_Type_Spec, "ver")) {
+	return -1;
+    }
+
+    if (!initAndAddType(m, &modstate->spec_Type, &spec_Type_Spec, "spec")) {
+	return -1;
+    }
+    if (!initAndAddType(m, &modstate->specPkg_Type, &specPkg_Type_Spec, "specPkg")) {
+	return -1;
+    }
 
     addRpmTags(m);
 
@@ -458,6 +477,13 @@ static int initModule(PyObject *m)
     REGISTER_ENUM(RPMFILE_PUBKEY);
     REGISTER_ENUM(RPMFILE_ARTIFACT);
 
+    REGISTER_ENUM(RPMFI_FLAGS_ERASE);
+    REGISTER_ENUM(RPMFI_FLAGS_INSTALL);
+    REGISTER_ENUM(RPMFI_FLAGS_VERIFY);
+    REGISTER_ENUM(RPMFI_FLAGS_QUERY);
+    REGISTER_ENUM(RPMFI_FLAGS_FILETRIGGER);
+    REGISTER_ENUM(RPMFI_FLAGS_ONLY_FILENAMES);
+
     REGISTER_ENUM(RPMDEP_SENSE_REQUIRES);
     REGISTER_ENUM(RPMDEP_SENSE_CONFLICTS);
 
@@ -484,6 +510,8 @@ static int initModule(PyObject *m)
     REGISTER_ENUM(RPMSENSE_KEYRING);
     REGISTER_ENUM(RPMSENSE_CONFIG);
     REGISTER_ENUM(RPMSENSE_MISSINGOK);
+    REGISTER_ENUM(RPMSENSE_PREUNTRANS);
+    REGISTER_ENUM(RPMSENSE_POSTUNTRANS);
 
     REGISTER_ENUM(RPMTRANS_FLAG_TEST);
     REGISTER_ENUM(RPMTRANS_FLAG_BUILD_PROBS);
@@ -496,10 +524,13 @@ static int initModule(PyObject *m)
     REGISTER_ENUM(RPMTRANS_FLAG_KEEPOBSOLETE);
     REGISTER_ENUM(RPMTRANS_FLAG_NOCONTEXTS);
     REGISTER_ENUM(RPMTRANS_FLAG_NOCAPS);
+    REGISTER_ENUM(RPMTRANS_FLAG_NODB);
     REGISTER_ENUM(RPMTRANS_FLAG_REPACKAGE);
     REGISTER_ENUM(RPMTRANS_FLAG_REVERSE);
     REGISTER_ENUM(RPMTRANS_FLAG_NOPRE);
     REGISTER_ENUM(RPMTRANS_FLAG_NOPOST);
+    REGISTER_ENUM(RPMTRANS_FLAG_NOPREUNTRANS);
+    REGISTER_ENUM(RPMTRANS_FLAG_NOPOSTUNTRANS);
     REGISTER_ENUM(RPMTRANS_FLAG_NOTRIGGERPREIN);
     REGISTER_ENUM(RPMTRANS_FLAG_NOTRIGGERIN);
     REGISTER_ENUM(RPMTRANS_FLAG_NOTRIGGERUN);
@@ -514,6 +545,7 @@ static int initModule(PyObject *m)
     REGISTER_ENUM(RPMTRANS_FLAG_ADDINDEPS);
     REGISTER_ENUM(RPMTRANS_FLAG_NOCONFIGS);
     REGISTER_ENUM(RPMTRANS_FLAG_DEPLOOPS);
+    REGISTER_ENUM(RPMTRANS_FLAG_NOARTIFACTS);
 
     REGISTER_ENUM(RPMPROB_FILTER_IGNOREOS);
     REGISTER_ENUM(RPMPROB_FILTER_IGNOREARCH);
@@ -584,8 +616,12 @@ static int initModule(PyObject *m)
     REGISTER_ENUM(RPMVSF_NEEDPAYLOAD);
     REGISTER_ENUM(RPMVSF_NOSHA1HEADER);
     REGISTER_ENUM(RPMVSF_NOSHA256HEADER);
+    REGISTER_ENUM(RPMVSF_NOSHA3_256HEADER);
     REGISTER_ENUM(RPMVSF_NODSAHEADER);
     REGISTER_ENUM(RPMVSF_NORSAHEADER);
+    REGISTER_ENUM(RPMVSF_NOSHA512PAYLOAD);
+    REGISTER_ENUM(RPMVSF_NOSHA3_256PAYLOAD);
+    REGISTER_ENUM(RPMVSF_NOSHA256PAYLOAD);
     REGISTER_ENUM(RPMVSF_NOPAYLOAD);
     REGISTER_ENUM(RPMVSF_NOMD5);
     REGISTER_ENUM(RPMVSF_NODSA);
@@ -669,6 +705,74 @@ static int initModule(PyObject *m)
     REGISTER_ENUM(RPMSPEC_FORCE);
     REGISTER_ENUM(RPMSPEC_NOLANG);
 
-    return 1;
+    return 0;
 }
 
+rpmmodule_state_t *rpmModState_FromModule(PyObject *m) {
+    assert(PyModule_GetDef(m) == &moduledef);
+    rpmmodule_state_t *state = PyModule_GetState(m);
+    if (state == NULL && !PyErr_Occurred()) {
+	PyErr_SetString(PyExc_SystemError, "could not get _rpm module state");
+    }
+    return state;
+}
+
+/* PyType_GetModuleByDef is only in Python 3.11+ */
+static PyObject *
+MyType_GetModuleByDef(PyTypeObject *type, PyModuleDef *def)
+{
+    assert(PyType_Check(type));
+    PyObject *m = PyType_GetModule(type);
+    if (m && PyModule_GetDef(m) == def) {
+	return m;
+    }
+    if (!m && PyErr_Occurred()) {
+	if (PyErr_ExceptionMatches(PyExc_TypeError)) {
+	    PyErr_Clear();
+	} else {
+	    return NULL;
+	}
+    }
+
+    PyObject *mro = PyObject_CallMethod((PyObject*)type, "mro", "");
+    if (!mro) {
+	return NULL;
+    }
+    size_t size = PyList_Size(mro);
+    for (Py_ssize_t i = 1; i < size; i++) {
+        PyObject *super = PyList_GetItem(mro, i); // borrowed reference
+	if (!super) {
+	    Py_DECREF(mro);
+	    return NULL;
+	}
+	m = PyType_GetModule((PyTypeObject*)super);
+	if (m && PyModule_GetDef(m) == def) {
+	    Py_DECREF(mro);
+	    return m;
+	}
+	if (!m && PyErr_Occurred()) {
+	    if (PyErr_ExceptionMatches(PyExc_TypeError)) {
+		PyErr_Clear();
+	    } else {
+		Py_DECREF(mro);
+		return NULL;
+	    }
+	}
+    }
+
+    PyErr_SetString(PyExc_SystemError, "could not get _rpm module state");
+    Py_DECREF(mro);
+    return NULL;
+}
+
+rpmmodule_state_t *rpmModState_FromType(PyTypeObject *t) {
+    PyObject *mod = MyType_GetModuleByDef(t, &moduledef);
+    if (!mod) {
+	return NULL;
+    }
+    return rpmModState_FromModule(mod);
+}
+
+rpmmodule_state_t *rpmModState_FromObject(PyObject *s) {
+    return rpmModState_FromType(Py_TYPE(s));
+}
